#include <sgjbase.h>
#include <sgjmig.h>

struct shot_data_par_type shot_data_init(){
  struct shot_data_par_type data_par;
  filehead data;
  int stat;
  stat=file_open_input("Data",&data);
  if (data.ndims<5){
    data_par.nsy=1; data_par.dsy=10.0; data_par.osy=0.0;
  }
  if (data.ndims==5)
  {
    data_par.nsy=data.n[4]; data_par.dsy=data.d[4]; data_par.osy=data.o[4];
  }
  if (data.ndims<4){
     data_par.nsx=1; data_par.dsx=10.0; data_par.osx=0.0;
  }
  else{
    data_par.nsx=data.n[3]; data_par.dsx=data.d[3]; data_par.osx=data.o[3];
  }
  data_par.nw= data.n[2]; data_par.dw= data.d[2]; data_par.ow =data.o[2];
  data_par.nhy=data.n[1]; data_par.dhy=data.d[1]; data_par.ohy=data.o[1];
  data_par.nhx=data.n[0]; data_par.dhx=data.d[0]; data_par.ohx=data.o[0];
  
  if (data_par.nsx==0) data_par.nsx=1;
  data_par.wavelet=allocatec(data_par.nw);
  data_par.tmp_wld=allocatec(data_par.nhx*data_par.nhy);
  return data_par;

}

void shot_data_release(struct shot_data_par_type *data_par){
  free(data_par->wavelet); free(data_par->tmp_wld);
}

// data wavelet should run after shot_data_init
void shot_data_wavelet(struct shot_data_par_type data_par){
  __complex__ float *wavelet=data_par.wavelet;
  filehead wavlt;
  int stat;
  stat=file_open_input("Wavelet",&wavlt);
  if ( wavlt.n[0] != data_par.nw ) seperr("wavelet should have the same frequency  as data");
  if ( sreed("Wavelet",wavelet,wavlt.n[0]*8)!=wavlt.n[0]*8 ) seperr("err read wavelet");
}

// data_1shot_init in f90
void shot_data_grab_mig_par(int i_sx,int i_sy,float *mig_x_min,float *mig_y_min, int *nx,int *ny,struct shot_data_par_type data_par )
{
  float sx,sy;
  sx=((float)i_sx)*data_par.dsx+data_par.osx; sy=((float) i_sy)*data_par.dsy+data_par.osy; 
  *mig_x_min=sx+data_par.ohx;                  *mig_y_min=sy+data_par.ohy;
  *nx=data_par.nhx;                           *ny=data_par.nhy;
}

void data_source(float min_x,float min_y,int nx,int ny,int iw,int i_sx,int i_sy,__complex__ float *wld,struct shot_data_par_type data_par)
{
  float sx,sy;
  int ix,iy;
  float data_dsx, data_osx, data_dsy, data_osy, data_dhx, data_dhy;
  __complex__ float * wavelet=data_par.wavelet;
  data_dsx=data_par.dsx; data_osx=data_par.osx; data_dsy=data_par.dsy; data_osy=data_par.osy;
  data_dhx=data_par.dhx; data_dhy=data_par.dhy;

  sx=i_sx*data_dsx+data_osx;               sy=i_sy*data_dsy+data_osy;
  ix=(sx-min_x)/data_dhx; iy=(sy-min_y)/data_dhy;
  if (ix <0 || iy <0 || ix>=nx || iy>=ny) seperr("source point is outside of wavefield area");
  vector_value_c(wld,cmplx(0.0,0.0),nx*ny);  //wld=0.0
  wld[i2(iy,ix,nx)]=wavelet[iw]; if (ix+1 <nx)   wld[i2(iy,ix+1,nx)]=0.5*wavelet[iw]; if (ix-1 >=0  ) wld[i2(iy,ix-1,nx)]=0.5*wavelet[iw];
}



void data_receiver(float min_x,float min_y,int nx,int ny,int iw,int i_sx,int i_sy,__complex__ float *wld,struct shot_data_par_type data_par)
{
  __complex__ float *tmp_wld=data_par.tmp_wld;
  float sx,sy,max_x,max_y,min_x_d,min_y_d,max_x_d,max_y_d;
  float max_x_read,max_y_read,min_x_read,min_y_read;
  int  nx_read,ny_read,i_ox_read,i_oy_read,i_ox_wld,i_oy_wld;
  int n_block,blocksize;
  int i_s,ix,iy,ix_wld,iy_wld,ix_read,iy_read;
  float data_dhx,data_dhy,data_ohx,data_ohy,data_osx,data_osy,data_dsx,data_dsy;
  int data_nhx,data_nhy,data_nsx,data_nw;
  data_dhx=data_par.dhx; data_ohx=data_par.ohx; data_nhx=data_par.nhx; 
  data_dhy=data_par.dhy; data_ohy=data_par.ohy; data_nhy=data_par.nhy; 
  data_osx=data_par.osx; data_dsx=data_par.dsx; data_nsx=data_par.nsx; 
  data_osy=data_par.osy; data_dsy=data_par.dsy;
  data_nw=data_par.nw;


  vector_value_c(wld,cmplx(0.0,0.0),nx*ny);  //wld=0.0
  sx=i_sx*data_dsx+data_osx;                    sy=i_sy*data_dsy+data_osy; 
  max_x=min_x+(nx-1)*data_dhx;                  max_y=min_y+(ny-1)*data_dhy;
  min_x_d=sx+data_ohx;                          min_y_d=sy+data_ohy;         
  max_x_d=min_x_d+data_dhx*(data_nhx-1);        max_y_d=min_y_d+data_dhy*(data_nhy-1);
  max_x_read=MIN(max_x_d,max_x);                max_y_read=MIN(max_y_d,max_y);
  min_x_read=MAX(min_x_d,min_x);                min_y_read=MAX(min_y_d,min_y);
  nx_read=(max_x_read-min_x_read)/data_dhx+1;   ny_read=(max_y_read-min_y_read)/data_dhy+1;
  i_ox_read=(min_x_read-min_x_d)/data_dhx;      i_oy_read=(min_y_read-min_y_d)/data_dhy;
  i_ox_wld=(min_x_read-min_x)/data_dhx;         i_oy_wld=(min_y_read-min_y)/data_dhy;
  i_s=i_sy*data_nsx+i_sx;                       n_block=i_s*data_nw+iw;                 
  blocksize=data_nhy*data_nhx*8;
  //printf("blocksize=%d  n_block=%d\n",blocksize,n_block);
  if ( sseek_block("Data",n_block,blocksize,0) !=n_block) seperr("data seek wrong");
  if ( sreed("Data",tmp_wld,blocksize )!= blocksize) seperr("data read wrong");
  for(iy=0;iy<ny_read;iy++){
    iy_wld=iy+i_oy_wld;iy_read=iy+i_oy_read;
    for(ix=0;ix<nx_read;ix++){
      ix_wld=ix+i_ox_wld;ix_read=ix+i_ox_read; 
      wld[i2(iy_wld,ix_wld,nx)]=tmp_wld[i2(iy_read,ix_read,data_nhx)];
    }
  }
  //wld(i_ox_wld:i_ox_wld+nx_read-1,i_oy_wld:i_oy_wld+ny_read-1)=tmp_wld(i_ox_read:i_ox_read+nx_read-1,i_oy_read:i_oy_read+ny_read-1);
}

void shot_data_plane_grab_mig_par(int i_sx,int i_sy,float *mig_x_min,float *mig_y_min,int *nx,int *ny,struct shot_data_par_type data_par,struct shot_plane_par_type *plane_par )
{
  int i_px,i_py;
  float px,py;
  float plane_osx,plane_dsx,plane_osy,plane_dsy;
  int   plane_nsx,plane_nsy;

  *mig_x_min=data_par.ohx;  *mig_y_min=data_par.ohy;  *nx=data_par.nhx;   *ny=data_par.nhy;
//  i_px=i_sx;  i_py=i_sy;   px=i_px*data_par.dsx+data_par.osx;     py=i_py*data_par.dsy+data_par.osy;
//sin_rotate_angle=px*1500.0
//rotate_angle=asin(sin_rotate_angle)
//if ( abs(rotate_angle) <10.0/180.0*3.1415926) then
//     rotate_angle=0.0
//end if
  if (0==auxpar("plane_os_x","f", &plane_osx,"Data")) plane_osx=0.0;
  if (0==auxpar("plane_ds_x","f", &plane_dsx,"Data")) plane_dsx=1.0;
  if (0==auxpar("plane_ns_x","d", &plane_nsx,"Data")) plane_nsx=1;
  if (0==auxpar("plane_os_y","f", &plane_osy,"Data")) plane_osy=0.0;
  if (0==auxpar("plane_ds_y","f", &plane_dsy,"Data")) plane_dsy=1.0;
  if (0==auxpar("plane_ns_y","d", &plane_nsy,"Data")) plane_nsy=1;
  
  plane_par->osx=plane_osx; plane_par->dsx=plane_dsx; plane_par->nsx=plane_nsx;
  plane_par->osy=plane_osy; plane_par->dsy=plane_dsy; plane_par->nsy=plane_nsy;
  plane_par->ohx=data_par.ohx; plane_par->dhx=data_par.dhx; plane_par->nhx=data_par.nhx;
  plane_par->ohy=data_par.ohy; plane_par->dhy=data_par.dhy; plane_par->nhy=data_par.nhy;
  //printf("!!!!!!!nsx=%d",plane_par->nsx);
}

void shot_plane_source(float min_x,float min_y,int nx,int ny,int iw,int i_px,int i_py,__complex__ float *wld,struct shot_data_par_type data_par,struct shot_plane_par_type plane_par)
{
  float x_dt,y_dt,t_x;
  int iy,ix,isx,isy;
  float plane_s_x,px,py,w;
  float plane_dx,plane_dw,plane_ow,plane_opx,plane_dpx,plane_opy,plane_dpy,plane_osx,plane_dsx,plane_osy,plane_dsy;
  int plane_nw,plane_npx,plane_npy,plane_nsx,plane_nsy;
  __complex__ float * wavelet=data_par.wavelet;
  //printf("@@@@@plane.nsx=%d",plane_par.nsx);
  plane_dw=data_par.dw; plane_ow=data_par.ow; 
  plane_dpx=data_par.dsx; plane_opx=data_par.osx; plane_npx=data_par.nsx;
  plane_dpy=data_par.dsy; plane_opy=data_par.osy; plane_npy=data_par.nsy;
  plane_nsx=plane_par.nsx; plane_dsx=plane_par.dsx; plane_osx=plane_par.osx;
  plane_nsy=plane_par.nsy; plane_dsy=plane_par.dsy; plane_osy=plane_par.osy;
  plane_dx= data_par.dhx;

  // data_par.dsx: plane_par.dpx;  
  w=(float)(iw)*plane_dw+plane_ow;
  px=(float)(i_px)*plane_dpx+plane_opx;     py=(float)(i_py)*plane_dpy+plane_opy;
  vector_value_c(wld,cmplx(0.0,0.0),nx*ny);  //wld=0.0
  x_dt=plane_dsx*px;   //plane_par.dsx is dsx of point source of original dataset
  y_dt=0.0;
  //printf("plane_nsy=%d,plane_nsx=%d\n",plane_nsy,plane_nsx);
  for(isy=0;isy<plane_nsy;isy++){  // plane_nsy  nsy of point source of original dataset
    iy=isy;
    for(isx=0;isx<plane_nsx;isx++){  // plane_nsx nsx of point source
        plane_s_x=(float)(isx)*plane_dsx+plane_osx;
        ix=(plane_s_x-min_x)/fabs(plane_dx);       t_x=(float)(isx)*x_dt;
        wld[i2(iy,ix,nx)]+=wavelet[iw]*cmplx(cos(w*t_x),sin(w*t_x));
    }
  }

}

void shot_plane_receiver(float min_x,float min_y,int nx,int ny,int iw,int i_sx,int i_sy,__complex__ float *wld,struct shot_data_par_type data_par,struct shot_plane_par_type plane_par)
{  // shot_plane_receiver

float  min_x_d,min_y_d,max_x_d,max_y_d,max_x,max_y;
float max_x_read,max_y_read,min_x_read,min_y_read;
int nx_read,ny_read,i_ox_read,i_oy_read,i_ox_wld,i_oy_wld,n_block,blocksize,i_s;
float plane_dx,plane_dy,plane_ox,plane_oy;
int plane_nx,plane_ny;
int ix,iy;
__complex__ float *tmp_wld=data_par.tmp_wld;

plane_dx=data_par.dhx;  plane_dy=data_par.dhy; plane_ox=data_par.ohx; plane_oy=data_par.ohy;
plane_nx=data_par.nhx;  plane_ny=data_par.nhy;


i_s=i_sy*data_par.nsx+i_sx;
max_x=min_x+(float)(nx-1)*fabs(plane_dx);              max_y=min_y+(float)(ny-1)*fabs(plane_dy);
min_x_d=plane_ox;                                      min_y_d=plane_oy;
max_x_d=min_x_d+fabs(plane_dx)*(plane_nx-1);           max_y_d=min_y_d+fabs(plane_dy)*(plane_ny-1);
max_x_read=MIN(max_x_d,max_x);                             max_y_read=MIN(max_y_d,max_y);
min_x_read=MAX(min_x_d,min_x);                             min_y_read=MAX(min_y_d,min_y);
nx_read=(max_x_read-min_x_read)/fabs(plane_dx)+1;      ny_read=(max_y_read-min_y_read)/fabs(plane_dy)+1;
i_ox_read=(min_x_read-min_x_d)/fabs(plane_dx);         i_oy_read=(min_y_read-min_y_d)/fabs(plane_dy);
i_ox_wld=(min_x_read-min_x)/fabs(plane_dx);            i_oy_wld=(min_y_read-min_y)/fabs(plane_dy);
n_block=i_s*data_par.nw+iw;        blocksize=plane_ny*plane_nx*8;
if ( sseek_block("Data",n_block,blocksize,0) !=n_block) seperr("data seek wrong");
if ( sreed("Data",tmp_wld,blocksize )!= blocksize) seperr("data read wrong");
for(iy=0;iy<ny_read;iy++)
  for(ix=0;ix<nx_read;ix++)
    wld[i2(i_oy_wld+iy,i_ox_wld+ix,nx)]=tmp_wld[i2(i_oy_read+iy,i_ox_read+ix,plane_nx)];

} // shot_plane_receiver




/*
__complex__ float * shot_data_wavelet_read(filehead *wavlt)
{
  int stat;
  __complex__ float *wavelet;
  stat=file_open_input("Wavelet",wavlt);
  printf("n is equal to %d",wavlt->n[0]);
  wavelet=allocatec(wavlt->n[0]);
  if ( sreed("Wavelet",wavelet,wavlt->n[0]*8) != wavlt->n[0]*8 ) seperr("err wavelet");
  return wavelet;
}

void shot_data_clear(__complex__ float *wavelet){
  free(wavelet);
}

void shot_data_source_modeling(float sx,float min_x,int nx,__complex__ float *wld,__complex__ float *wavelet, int iw, float dhx)
{

  int ix,iy;
  int ny;
  int iix,iiy;
  ny=1;iy=0;  ix=(sx-min_x)/dhx;
  if (ix <0 || ix>=nx ) seperr("err create data_source");

  for ( iiy=0;iiy < ny; iiy++){
    for (iix=0;iix<nx; iix++){
      wld[i2(iiy,iix,nx)]=cmplx(0.0,0.0);
    }
  }
  wld[i2(iy,ix,nx)]=wavelet[iw];
  if (ix+1 <=nx) wld[i2(iy,ix+1,nx)]=0.5*wavelet[iw];
  if (ix-1 >0  ) wld[i2(iy,ix-1,nx)]=0.5*wavelet[iw];
}
*/
