#include <math.h>
#include "sgjbase.h"
#include "sgjmig.h"

struct shot_ker_par_type shot_ker_p_init(int op,int nx,int ny,float dx,float dy,float dz,float a,float b,float trick,int is2d){
  struct shot_ker_par_type ker_par;


  float pi=3.1415926; int i;
  ker_par.nref=3;
  ker_par.op=op;
  ker_par.nx=nx; ker_par.ny=ny;
  ker_par.dx=dx; ker_par.dy=dy; ker_par.dz=dz;
  ker_par.fkx=-pi/ker_par.dx; ker_par.fky=-pi/ker_par.dy;  //change tilt 
  ker_par.dkx=2.0*pi/(nx*ker_par.dx); ker_par.dky=2.0*pi/(ny*ker_par.dy);  //change in tilt
  ker_par.ax=a*dz/(2.0*ker_par.dx*ker_par.dx); ker_par.ay=a*dz/(2.0*ker_par.dy*ker_par.dy);
  ker_par.bx=b/(ker_par.dx*ker_par.dx); ker_par.by=b/(ker_par.dy*ker_par.dy);
  ker_par.tx=trick;     ker_par.ty=trick;
  ker_par.is2d=is2d;
  ker_par.ux=allocatec(nx);  ker_par.vx=allocatec(nx); 
  ker_par.cax=allocatef(nx); ker_par.cbx=allocatef(nx);
  ker_par.uy=allocatec(ny);  ker_par.vy=allocatec(ny); 
  ker_par.cay=allocatef(ny); ker_par.cby=allocatef(ny);
  ker_par.ra=allocatec(nx);  ker_par.rb=allocatec(nx); ker_par.rc=allocatec(nx);
  ker_par.tmp_fft=allocatec(nx); ker_par.add_wld=allocatec(nx*ny); ker_par.tmp_wld=allocatec(nx*ny);
  ker_par.v_zw=allocatef(nx*ny); ker_par.v_zw0=allocatef(nx*ny);
  ker_par.v0_z=allocatef(ker_par.nref);
  ker_par.Nsin=2000;
  ker_par.sin=allocatef(ker_par.Nsin); 
  ker_par.cos=allocatef(ker_par.Nsin); 
  ker_par.sqrt=allocatef(ker_par.Nsin);
  ker_par.dsin=pi*2.0/(float)(ker_par.Nsin);
  ker_par.dqrt=1.0/(float)(ker_par.Nsin);
  for (i=0;i<ker_par.Nsin;i++){
    ker_par.sin[i]=sin(ker_par.dsin*i);
    ker_par.cos[i]=cos(ker_par.dsin*i);
    ker_par.sqrt[i]=sqrt(ker_par.dqrt*i);
  }
  
  return ker_par;
}



void shot_ker_release(struct shot_ker_par_type *ker_par){
  free(ker_par->ux); free(ker_par->vx); free(ker_par->cax); free(ker_par->cbx);
  free(ker_par->uy); free(ker_par->vy); free(ker_par->cay); free(ker_par->cby);
  free(ker_par->ra); free(ker_par->rb); free(ker_par->rc); 
  free(ker_par->tmp_fft);free(ker_par->add_wld); free(ker_par->tmp_wld);
  free(ker_par->v_zw); free(ker_par->v_zw0); free(ker_par->v0_z);
  free(ker_par->sin); free(ker_par->cos); free(ker_par->sqrt);
}

void tilt_nx_dkx_change_iz(struct shot_ker_par_type *ker_par, int nx_iz, float dkx_iz)
{
    ker_par->dkx=dkx_iz; ker_par->nx=nx_iz;
}

void shot_ker_depth_onestep(__complex__ float *wld_z,float *v_z,float w,int signn, struct shot_ker_par_type ker_par){
int op,nref,nx,ny,is2d;

float wvx,dref,wvx_all;
__complex__ float  *tmp_fft,*add_wld,*tmp_wld;
int iref,ix,iy,ivx,iiref;
float w_v0_z,maxvz,minvz;
float *v0_z,*v_zw,*v_zw0;
nref=ker_par.nref;  nx=ker_par.nx;  ny=ker_par.ny;  is2d=ker_par.is2d;  op=ker_par.op; //! op:1:ssf 2:fd 3:ffd
v0_z=ker_par.v0_z; v_zw=ker_par.v_zw; v_zw0=ker_par.v_zw0;
tmp_fft=ker_par.tmp_fft; add_wld=ker_par.add_wld; tmp_wld=ker_par.tmp_wld;
//!v0_z(1)=minval(v_z);//!v0_z(nref)=maxval(v_z);//!dref=(v0_z(nref)-v0_z(1))/(nref-1)

maxvz=maxval(v_z,nx*ny); minvz=minval(v_z,nx*ny);  dref=(maxvz-minvz)/(nref+1);

if ( fabs(maxvz-minvz)<50.0 ) { ///?????????????wait
   nref=1;  v0_z[0]=0.5*(maxvz+minvz);
}
else{
   v0_z[0]=minvz;  v0_z[nref-1]=maxvz;  dref=(v0_z[nref-1]-v0_z[0])/(float)(nref-1); v0_z[1]=(minvz+maxvz)*0.5;  
   //v0_z(2)=(minval(v_z)+maxval(v_z))*0.5; //!  do iref=1,nref //!    v0_z(iref)=iref*dref //!  end do
}

//!v0_z(2)=0.5*v0_z(1)+0.5*v0_z(3);//!v0_z(1)=0.5*v0_z(1)+0.5*v0_z(2);//!v0_z(3)=0.5*v0_z(3)+0.5*v0_z(2)
//v0_z[0]=0.5*(v0_z[0]+v0_z[1]); v0_z[2]=0.5*(v0_z[1]+v0_z[2]); dref=v0_z[1]-v0_z[0];
//nref=1;  v0_z[0]=minvz; 
for (iy=0;iy<ny;iy++)
  for(ix=0;ix<nx;ix++)
    v_zw[i2(iy,ix,nx)]=v_z[i2(iy,ix,nx)]/w;
if (op == 2){
   w_v0_z=0.0;
   shot_ker_ssf(wld_z,w_v0_z,w,v_z,ker_par,signn);
   shot_ker_fd(wld_z,v_zw,v_zw,ker_par,signn);
}
else{
  if (is2d){
    rowc(wld_z,tmp_fft,0,nx,1);  //tmp_fft=wld_z(1,:);
    cwpfft1d(tmp_fft,nx,+1);
    rowc(wld_z,tmp_fft,0,nx,0);  //wld_z(1,:)=tmp_fft(:);
  }
  else{
    for(iy=0;iy<ny;iy++){
      rowc(wld_z,tmp_fft,iy,nx,1);  //tmp_fft=wld_z(1,:);
      cwpfft1d(tmp_fft,nx,+1);
      rowc(wld_z,tmp_fft,iy,nx,0);  //wld_z(1,:)=tmp_fft(:);
    }
    for(ix=0;ix<nx;ix++){
      colc(wld_z,tmp_fft,ix,nx,ny,1);
      cwpfft1d(tmp_fft,ny,+1);
      colc(wld_z,tmp_fft,ix,nx,ny,0);   // wait tmp_fftx tmp_ffty
    }
  
  }
  
  
  for( iy=0;iy<ny;iy++)
    for(ix=0;ix<nx;ix++)
       add_wld[i2(iy,ix,nx)]=cmplx(0.0,0.0);   //add_wld=0.0;

  for( iref=0;iref<nref;iref++){
    matrix_equ_c(tmp_wld,wld_z,ny,nx);//tmp_wld=wld_z;
    w_v0_z=w/v0_z[iref];
    shot_phase_shift(tmp_wld,w_v0_z,ker_par,signn);
    if (is2d){
      rowc(tmp_wld,tmp_fft,0,nx,1); //tmp_fft=tmp_wld(:,1);
      cwpfft1d(tmp_fft,nx,-1);
      rowc(tmp_wld,tmp_fft,0,nx,0);  //tmp_wld(:,1)=tmp_fft(:);
    }
    else{
      for(ix=0;ix<nx;ix++){
        colc(tmp_wld,tmp_fft,ix,nx,ny,1);
        cwpfft1d(tmp_fft,ny,-1);
        colc(tmp_wld,tmp_fft,ix,nx,ny,0);   // wait tmp_fftx tmp_ffty
      }
      for(iy=0;iy<ny;iy++){
        rowc(tmp_wld,tmp_fft,iy,nx,1);  //tmp_fft=wld_z(1,:);
        cwpfft1d(tmp_fft,nx,-1);
        rowc(tmp_wld,tmp_fft,iy,nx,0);  //wld_z(1,:)=tmp_fft(:);
      }

    }
    shot_ker_ssf(tmp_wld,w_v0_z,w,v_z,ker_par,signn);
    if (op==3){ 
      for( iy=0;iy<ny;iy++)
         for(ix=0;ix<nx;ix++)
             v_zw0[i2(iy,ix,nx)]=v_zw[i2(iy,ix,nx)]-v0_z[iref]/w;
      shot_ker_fd(tmp_wld,v_zw0,v_zw,ker_par,signn);
    }
    if ( nref==1) vector_cp_c(add_wld,tmp_wld,ny*nx); //add_wld=tmp_wld;
    else{
      for(iy=0;iy<ny;iy++){      
        for(ix=0;ix<nx;ix++){  
          ivx=(int)((v_z[i2(iy,ix,nx)]-v0_z[0])/dref); //wait ???
          wvx=(v_z[i2(iy,ix,nx)]-(v0_z[0]+(float)(ivx)*dref))/dref;   //wait ???
          if (v_z[i2(iy,ix,nx)] >= v0_z[nref-1]){
            if (iref==nref-1) add_wld[i2(iy,ix,nx)]+=tmp_wld[i2(iy,ix,nx)];
          }
          else{
            if (v_z[i2(iy,ix,nx)] <= v0_z[0]){ 
              if (iref==0)  add_wld[i2(iy,ix,nx)]+=tmp_wld[i2(iy,ix,nx)];
            }
            else{
              if (ivx==iref) add_wld[i2(iy,ix,nx)] += tmp_wld[i2(iy,ix,nx)]*(1-wvx);
              if (ivx+1==iref) add_wld[i2(iy,ix,nx)] += tmp_wld[i2(iy,ix,nx)]*wvx;
            }
          }
        }
      }
    }
    
  }
  vector_cp_c(wld_z,add_wld,ny*nx);  //wld_z=add_wld;
}

}


void shot_phase_shift(__complex__ float *wld_z,float w_v0_z,struct shot_ker_par_type ker_par,int signn)
{// wld_z(ker_par.ny,ker_par.nx)
  int nx,ny;
  float fkx,fky,dkx,dky,dz;
  
  int ix,iy,ipi,isin,ikz2;
  float kx,ky,kz,kz2,phsft,pio2,wpi,tphsft,wsin,sin_phsft,cos_phsft;
  pio2=3.1415926*2.0;
  nx=ker_par.nx; ny=ker_par.ny; 
  fkx=ker_par.fkx; fky=ker_par.fky; 
  dkx=ker_par.dkx; dky=ker_par.dky;
  dz=ker_par.dz;

  for(iy=0;iy<ny;iy++){ //do iy=1,ny
    if (ker_par.is2d) 
      ky=0.0;
    else
      ky=fky+iy*dky;
 
    for(ix=0;ix<nx;ix++){  //do ix=1,nx
      kx=fkx+ix*dkx;
      kz2=(kx*kx+ky*ky)/(w_v0_z*w_v0_z);
      if (kz2 <=1.0) {
        kz2=1.0-kz2;
        //ikz2=(int)(kz2/ker_par.dqrt);
        //phsft=-signn*dz*w_v0_z*(ker_par.sqrt[ikz2]);
        phsft=-signn*dz*w_v0_z*(sqrt(kz2));
        /*
        tphsft=fabs(phsft); 
        ipi=(int)(tphsft/pio2); wpi=tphsft-(float)(ipi)*pio2;
        isin=(int)(wpi/ker_par.dsin); //wsin=wpi-(float)(isin)*ker_par.dsin;
        sin_phsft=ker_par.sin[isin];//*(1.0-wsin)+ker_par.sin[isin+1]*wsin;
        cos_phsft=ker_par.cos[isin];//*(1.0-wsin)+ker_par.sin[isin+1]*wsin;
        if (phsft < 0){
          sin_phsft=-sin_phsft;
        }
        wld_z[i2(iy,ix,nx)]=wld_z[i2(iy,ix,nx)]*cmplx(cos_phsft,sin_phsft);
        */
        wld_z[i2(iy,ix,nx)]=wld_z[i2(iy,ix,nx)]*cmplx(cos(phsft),sin(phsft));
      }
      else{
        if (kz2>2) kz2=2.0;
        kz2=kz2-1;
        ikz2=(int)(kz2/ker_par.dqrt);
        //phsft=dz*w_v0_z*(ker_par.sqrt[ikz2]);
        phsft=dz*w_v0_z*(sqrt(kz2));
        wld_z[i2(iy,ix,nx)]=wld_z[i2(iy,ix,nx)]*exp(-phsft);
      }     
    } 
  }
}


void shot_ker_ssf(__complex__ float *wld_z,float w_v0_z,float w,float *v_z, struct shot_ker_par_type ker_par,int signn)
{ // wld_z(ny,nx) v_z(ny,nx)
  int nx,ny,ix,iy;
  float phsft,dz;
  int ipi,isin;
  float pio2,wpi,tphsft,wsin,sin_phsft,cos_phsft;
  pio2=3.1415926*2.0;

  nx=ker_par.nx; ny=ker_par.ny; dz=ker_par.dz;
  
  for(iy=0;iy<ny;iy++){
    for(ix=0;ix<nx;ix++){
      phsft=-signn*dz*(w/v_z[i2(iy,ix,nx)]-w_v0_z);
      //tphsft=fabs(phsft); 
      //ipi=(int)(tphsft/pio2); wpi=tphsft-(float)(ipi)*pio2;
      //isin=(int)(wpi/ker_par.dsin); //wsin=wpi-(float)(isin)*ker_par.dsin;
      //sin_phsft=ker_par.sin[isin];//*(1.0-wsin)+ker_par.sin[isin+1]*wsin;
      //cos_phsft=ker_par.cos[isin];//*(1.0-wsin)+ker_par.sin[isin+1]*wsin;
      //if (phsft < 0){
      //  sin_phsft=-sin_phsft;
      //}
      //wld_z[i2(iy,ix,nx)]=wld_z[i2(iy,ix,nx)]*cmplx(cos_phsft,sin_phsft);
      wld_z[i2(iy,ix,nx)]=wld_z[i2(iy,ix,nx)]*cmplx(cos(phsft),sin(phsft));
    }
  }
}

void shot_ker_fd(__complex__ float *wld_z,float *ca,float *cb,struct shot_ker_par_type ker_par,int signn)
{
//complex :: wld_z(:,:)
//real    :: ca(:,:),cb(:,:)

  int nx,ny,ix,iy;
  float tx,ty,ax,ay,bx,by;
  __complex__ float *ux,*vx,*uy,*vy;
  float *cax,*cbx,*cay,*cby;

  ax=-signn*ker_par.ax; ay=-signn*ker_par.ay; 
  bx=ker_par.bx; by=ker_par.by; 
  nx=ker_par.nx; ny=ker_par.ny;
  tx=ker_par.tx; ty=ker_par.ty;

  ux=ker_par.ux; vx=ker_par.vx; cax=ker_par.cax; cbx=ker_par.cbx;
  uy=ker_par.uy; vy=ker_par.vy; cay=ker_par.cay; cby=ker_par.cby;

  //  
  //!do ix=1,nx
  //!  uy=wld_z(ix,:)
  //!  cay=ca(ix,:)
  //!  cby=cb(ix,:)
  //!  call fd(uy,vy,ty,ay,by,cay,cby,ny,ray,rby,rcy)
  //!  wld_z(ix,:)=vy
  //!end do
  //!write(*,*) "maxval",maxval(real(wld_z))

  for(iy=0;iy<ny;iy++){
    rowc(wld_z,ux,iy,nx,1);  //ux=wld_z(iy,:)
    rowf(ca,cax,iy,nx,1);   //  cax=ca(iy,:)
    rowf(cb,cbx,iy,nx,1);   //  cbx=cb(iy,:)
    fd(ux,vx,tx,ax,bx,cax,cbx,nx,ker_par); //,rax,rbx,rcx);
    rowc(wld_z,vx,iy,nx,0);//  wld_z(iy,:,:)=vx
  }
//  free(ux); free(vx); free(cax); free(cbx); 
//  free(uy); free(vy); free(cay); free(cby);
}


void fd(__complex__ float *u,__complex__ float *v,float tr,float a,float b,float *ca,float *cb,int nx,struct shot_ker_par_type ker_par)
{ // u(nx),v(nx),ca(nx),cb(nx)

  __complex__ float *ra,*rb,*rc;
  __complex__ float  url,urr,ror,rol;
  int  ix;

  ra=ker_par.ra; rb=ker_par.rb; rc=ker_par.rc;
  for(ix=0;ix<nx;ix++){
    if (ix==0)    url=cmplx(0.0,0.0);
    else          url=u[ix-1];
    if (ix==nx-1) urr=cmplx(0.0,0.0);
    else          urr=u[ix+1];
    ror=cmplx(tr+b*cb[ix]*cb[ix],a*ca[ix]);
    rol=conjg(ror);
    v[ix]=u[ix]+ror*(url+urr-2.0*u[ix]);
    ra[ix]=1.0-2.0*rol;
    if (ix <nx-1) rb[ix] =rol;
    if (ix >0) rc[ix-1]=rol;
  }
  thryj(ra,rb,rc,v,nx,1);
//  free(ra); free(rb); free(rc);
}


void shot_ker_depth_onestep_ani_phaseshift(__complex__ float *wld_z,float *v_z,float w,float ep,float dl,float phi, float f,int signn, struct shot_ker_par_type ker_par){
int op,nref,nx,ny,is2d;

float wvx,dref,wvx_all;
__complex__ float  *tmp_fft,*add_wld,*tmp_wld;
int iref,ix,iy,ivx,iiref;
float w_v0_z,maxvz,minvz;
float *v0_z,*v_zw,*v_zw0;
nref=ker_par.nref;  nx=ker_par.nx;  ny=ker_par.ny;  is2d=ker_par.is2d;  op=ker_par.op; //! op:1:ssf 2:fd 3:ffd
v0_z=ker_par.v0_z; v_zw=ker_par.v_zw; v_zw0=ker_par.v_zw0;
tmp_fft=ker_par.tmp_fft; add_wld=ker_par.add_wld; tmp_wld=ker_par.tmp_wld;

maxvz=maxval(v_z,nx*ny); minvz=minval(v_z,nx*ny);  dref=(maxvz-minvz)/(nref+1);

if ( fabs(maxvz-minvz)<50.0 ) { ///?????????????wait
   nref=1;  v0_z[0]=0.5*(maxvz+minvz);
}
else{
   v0_z[0]=minvz;  v0_z[nref-1]=maxvz;  dref=(v0_z[nref-1]-v0_z[0])/(float)(nref-1); v0_z[1]=(minvz+maxvz)*0.5;  
}

for (iy=0;iy<ny;iy++)
  for(ix=0;ix<nx;ix++)
    v_zw[i2(iy,ix,nx)]=v_z[i2(iy,ix,nx)]/w;
if (op == 2){
   w_v0_z=0.0;
   shot_ker_ssf(wld_z,w_v0_z,w,v_z,ker_par,signn);
   shot_ker_fd(wld_z,v_zw,v_zw,ker_par,signn);
}
else{
  if (is2d){
    rowc(wld_z,tmp_fft,0,nx,1);  //tmp_fft=wld_z(1,:);
    cwpfft1d(tmp_fft,nx,+1);
    rowc(wld_z,tmp_fft,0,nx,0);  //wld_z(1,:)=tmp_fft(:);
  }
  else{
    for(iy=0;iy<ny;iy++){
      rowc(wld_z,tmp_fft,iy,nx,1);  //tmp_fft=wld_z(1,:);
      cwpfft1d(tmp_fft,nx,+1);
      rowc(wld_z,tmp_fft,iy,nx,0);  //wld_z(1,:)=tmp_fft(:);
    }
    for(ix=0;ix<nx;ix++){
      colc(wld_z,tmp_fft,ix,nx,ny,1);
      cwpfft1d(tmp_fft,ny,+1);
      colc(wld_z,tmp_fft,ix,nx,ny,0);   // wait tmp_fftx tmp_ffty
    }
  
  }

  for( iy=0;iy<ny;iy++)
    for(ix=0;ix<nx;ix++)
       add_wld[i2(iy,ix,nx)]=cmplx(0.0,0.0);   //add_wld=0.0;

  for( iref=0;iref<nref;iref++){
    matrix_equ_c(tmp_wld,wld_z,ny,nx);//tmp_wld=wld_z;
    w_v0_z=w/v0_z[iref];
    shot_phase_shift_ani(tmp_wld,w_v0_z,ep,dl,phi,f,ker_par,signn);
    if (is2d){
      rowc(tmp_wld,tmp_fft,0,nx,1); //tmp_fft=tmp_wld(:,1);
      cwpfft1d(tmp_fft,nx,-1);
      rowc(tmp_wld,tmp_fft,0,nx,0);  //tmp_wld(:,1)=tmp_fft(:);
    }
    else{
      for(ix=0;ix<nx;ix++){
        colc(tmp_wld,tmp_fft,ix,nx,ny,1);
        cwpfft1d(tmp_fft,ny,-1);
        colc(tmp_wld,tmp_fft,ix,nx,ny,0);   // wait tmp_fftx tmp_ffty
      }
      for(iy=0;iy<ny;iy++){
        rowc(tmp_wld,tmp_fft,iy,nx,1);  //tmp_fft=wld_z(1,:);
        cwpfft1d(tmp_fft,nx,-1);
        rowc(tmp_wld,tmp_fft,iy,nx,0);  //wld_z(1,:)=tmp_fft(:);
      }

    }
    //shot_ker_ssf(tmp_wld,w_v0_z,w,v_z,ker_par,signn);
    if (op==3){ 
      for( iy=0;iy<ny;iy++)
         for(ix=0;ix<nx;ix++)
             v_zw0[i2(iy,ix,nx)]=v_zw[i2(iy,ix,nx)]-v0_z[iref]/w;
      shot_ker_fd(tmp_wld,v_zw0,v_zw,ker_par,signn);
    }
    if ( nref==1) vector_cp_c(add_wld,tmp_wld,ny*nx); //add_wld=tmp_wld;
    else{
      for(iy=0;iy<ny;iy++){      
        for(ix=0;ix<nx;ix++){  
          ivx=(int)((v_z[i2(iy,ix,nx)]-v0_z[0])/dref); //wait ???
          wvx=(v_z[i2(iy,ix,nx)]-(v0_z[0]+(float)(ivx)*dref))/dref;   //wait ???
          if (v_z[i2(iy,ix,nx)] >= v0_z[nref-1]){
            if (iref==nref-1) add_wld[i2(iy,ix,nx)]+=tmp_wld[i2(iy,ix,nx)];
          }
          else{
            if (v_z[i2(iy,ix,nx)] <= v0_z[0]){ 
              if (iref==0)  add_wld[i2(iy,ix,nx)]+=tmp_wld[i2(iy,ix,nx)];
            }
            else{
              if (ivx==iref) add_wld[i2(iy,ix,nx)] += tmp_wld[i2(iy,ix,nx)]*(1-wvx);
              if (ivx+1==iref) add_wld[i2(iy,ix,nx)] += tmp_wld[i2(iy,ix,nx)]*wvx;
            }
          }
        }
      }
    }
    
  }

  vector_cp_c(wld_z,add_wld,ny*nx);  //wld_z=add_wld;
}

}


