/*
  Guojian Shan 
  email: shan@sep.stanford.edu
  This program creates explicit table for explicit method.
  Input file: Velocity, Ep, Dl, Wavelet
  Input parameter: maxrotateangle=45  dx=20 dz=4.0 vsvp=1.0-10^2/27^2 dep=0.01 ddl=0.01 dwvp=0.001
                   cal=1 (cal table or not) mkx=150 nft=5
  run example: explicittable.x Velocity=cellvelocity.H Wavelet=wavelet_exxon1.H Eps=cellep.H Dlt=celldl.H Extable=table.H
*/
#include <sepcube.h>
#include <sgjbase.h>
#include <sgjmig.h>

int main(int argc, char **argv)
{
  int mkx,nft,cal,stat,n_dl,n_ep,n_w_vp;
  float dx,dz,f,d_ep,d_dl,d_w_vp,o_ep,o_dl,o_w_vp,dkx,minvel,maxvel,minep,maxep,mindl,maxdl,minw,maxw;
  filehead velocity,eps,dlt,wavlt,table;
  __complex__ float *contablepp, *contablepm;

  int iangle,nangle; 
  float oangle,dangle,maxrotateangle,minrotateangle,phi; 

  initpar(argc,argv);
  
  if (0==getch("maxrotateangle","f",&maxrotateangle))  maxrotateangle=74.0;
  if (0==getch("dangle","f",&dangle) ) dangle=2.5;
  if (0==getch("mkx","d",&mkx)) mkx=150;
  if (0==getch("nft","d",&nft)) nft=19;
  if (0==getch("dx","f",&dx) ) dx=20.0;
  if (0==getch("dz","f",&dz) ) dz=4.0;
  if (0==getch("vsvp","f",&f) ) f=1.0-10.0*10.0/(27.0*27.0);
  if (0==getch("calcutable","d",&cal) ) cal=1;
  if (0==getch("dep","f",&d_ep) ) d_ep=0.01;
  if (0==getch("ddl","f",&d_dl) ) d_dl=0.01;
  if (0==getch("dwvp","f",&d_w_vp) ) d_w_vp=0.001;
  
  dkx=2.0*3.1415926/( (float)(2*mkx)*dx);  
  stat=file_open_input("Wavelet",&wavlt);
  minw=wavlt.o[0]; maxw=(float)(wavlt.n[0]-1)*wavlt.d[0];
  velocity_init(&velocity);  velocity_ani_init(&eps,&dlt);
  velocity_minmax("Velocity",velocity,&minvel,&maxvel); 
  velocity_minmax("Eps",eps,&minep,&maxep); velocity_minmax("Dlt",dlt,&mindl,&maxdl);
  printf("maxdl=%f\n",maxdl);
  o_ep=minep; n_ep=(maxep-o_ep)/d_ep+2;  o_dl=mindl; n_dl=(maxdl-o_dl)/d_dl+2; o_w_vp=minw/maxvel; n_w_vp=(maxw/minvel-o_w_vp)/d_w_vp+2;
  printf("o_ep=%f,n_ep=%d,o_dl=%f,n_dl=%d,o_w_vp=%f,n_w_vp=%d\n",o_ep, n_ep,o_dl,n_dl,o_w_vp,n_w_vp);
  contablepp=allocatec(nft*n_dl*n_ep*n_w_vp);
  contablepm=allocatec(nft*n_dl*n_ep*n_w_vp);
 

  nangle=maxrotateangle/dangle; maxrotateangle=nangle*dangle;
  oangle=0.0; nangle=(maxrotateangle-oangle)/dangle+1;
  printf("nangle=%d,oangle=%f\n",nangle,oangle);
  
  
  table.ndims=5;
  table.n[0]=2*nft;   table.d[0]=1.0;     table.o[0]=0.0;
  table.n[1]=n_dl;    table.d[1]=d_dl;    table.o[1]=o_dl;
  table.n[2]=n_ep;    table.d[2]=d_ep;    table.o[2]=o_ep;
  table.n[3]=n_w_vp;  table.d[3]=d_w_vp;  table.o[3]=o_w_vp;
  table.n[4]=nangle;  table.d[4]=dangle;  table.o[4]=oangle;
  stat=file_open_output_c("Extable",table);
 
  if (cal){
    for(iangle=0;iangle<nangle;iangle++){
      phi=(float)iangle*dangle+oangle;
      printf("tilting angle=%f\n",phi);
      phi=phi/180.0*3.1415926;
      explicit_table_old(contablepp,contablepm,phi,f,dx,dkx,dz,mkx,nft,o_ep,d_ep,n_ep,o_dl,d_dl,n_dl,o_w_vp,d_w_vp,n_w_vp);
      if (srite("Extable",contablepp,nft*n_dl*n_ep*n_w_vp*8)!=nft*n_dl*n_ep*n_w_vp*8 ) seperr("err rite table");
      if (srite("Extable",contablepm,nft*n_dl*n_ep*n_w_vp*8)!=nft*n_dl*n_ep*n_w_vp*8 ) seperr("err rite table");  
    }
  }
  
  free(contablepp); free(contablepm);
  return 0;
}
