/*
   Guojian Shan
   email: shan@sep.stanford.edu
   This program runs plane migration in VTI media in tilted coordinates.
   Input File: Data=data.H Velocity=velocity.H Wavelet=wavelet.H Eps=ep.H(\epsilon) Dlt=dl.H(\delta)
   Input parameters: nhx=  mig_nz=  mig_dz= image_min_x=  image_max_x=
   Output File: Image_hx=image_h.H (horizontal ODCIG) Image_hz=image_v.H (vertical ODCIG)
   ---------------------------------------------------------------------------------------------
   run: tiltplanemig.x Data=celldata.H Velocity=cellvelocity.H Wavelet=wavelet_exxon1.H 
                          Eps=cellep.H Dlt=celldl.H Extable=table.H
                          Image_hx=imagehx.H Image_hz=hz.H 
                          nhx=1 mig_nz=1000 mig_dz=4 image_min_x=-1000 image_max_x=9000.0 
   ----------------------------------------------------------------------------------------------
*/
#include <math.h>
#include <sepcube.h>
#include <sgjbase.h>
#include <sgjmig.h>

void mig_grab_par(int *op,int *mig_nz,float *mig_dz,float *fd_a,float *fd_b,float *trick,int *tilt);
void veltranspose(float *vel,int ny,int nx,int nz);
void getimagez(float *image,float *image_z,int iz,int ny,int nx,int nhy,int nhx,int nz);

int main(int argc, char **argv)
{
  struct shot_data_par_type   data_par;
  struct shot_plane_par_type   plane_par;
  struct shot_image_par_type  image_par;
  struct shot_ker_par_type    ker_par;
  struct shot_rotate_par_type rotate_par;
  filehead velocity,wave,eps,dlt;
  filehead rotimage,rotz;
  float mig_dx,mig_dy,mig_dz,mig_min_x,mig_min_y,fd_a,fd_b,trick,w;
  int op,mig_nz,mig_nx,mig_ny,tmp_mig_nx;
  int mig_nx_rotate,mig_nz_rotate,mig_ny_rotate;
  float rotate_angle,px;

  int i_sy,i_sx,iw,iz;
  float *vel,*ep,*dl,*image_z,*image,*image_rotate,*v_z,*ep_z,*dl_z;
  __complex__ float *wld_s,*wld_r,*wld_s_z0,*wld_r_z0,*wld_s_tilt,*wld_r_tilt;
  int nvel[2],ntmpvel[2];
  int data_nsy, data_nsx,ix,iy;
  int mig_size,image_size; int stat; int tilt;
  float dkx; int bx,ex,nx_iz,boundary_nx;

  initpar(argc,argv);


  velocity_init(&velocity);
  data_par=shot_data_init();
  printf("nsy=%d  nsx=%d\n",data_par.nsy,data_par.nsx); 
  shot_data_wavelet(data_par);
  image_par=shot_image_grab_par();

  mig_dx=data_par.dhx; mig_dy=data_par.dhy;
  mig_grab_par(&op,&mig_nz,&mig_dz,&fd_a,&fd_b,&trick,&tilt);
  shot_image_init(mig_dx,mig_dy,mig_dz,mig_nz,&image_par); 
  wave.d[1]=mig_dx; wave.d[2]=mig_dy; wave.d[0]=mig_dz; 
  
  data_nsy=data_par.nsy; data_nsx=data_par.nsx;

  for(i_sy=0;i_sy< data_nsy;i_sy++){
    for(i_sx=0;i_sx< data_nsx;i_sx++){
    //for(i_sx=0;i_sx< 1;i_sx++){
      shot_data_plane_grab_mig_par(i_sx,i_sy,&mig_min_x,&mig_min_y,&tmp_mig_nx,&mig_ny,data_par,&plane_par );
      mig_nx=tmp_mig_nx;
      printf("nx---------------%d\n",mig_nx); 
      rotate_angle=asin( (i_sx*data_par.dsx+data_par.osx)*1500.0)*1.6;
      //rotate_angle=-45.0/180.0*3.1415926;
      if (!tilt) rotate_angle=0.0;
      printf("rotate2=%f\n",rotate_angle/3.1415926*180.0);
      
      rotate_par=shot_rotate_init(rotate_angle,tmp_mig_nx,mig_nz,mig_dx,mig_dz);
      printf("nx=%d,nz=%d,ny=%d\n",rotate_par.nx_rotate,rotate_par.nz_rotate,rotate_par.ny_rotate);
      mig_nx_rotate=rotate_par.nx_rotate; mig_ny_rotate=rotate_par.ny_rotate;
      mig_nz_rotate=rotate_par.nz_rotate;
      
      mig_size=mig_nx_rotate*mig_ny_rotate; image_size=mig_size*image_par.nhy*image_par.nhx*mig_nz_rotate;

      ker_par=shot_ker_p_init(op,mig_nx_rotate,mig_ny_rotate,mig_dx,mig_dy,mig_dz,fd_a,fd_b,trick,1);
      vel=allocatef(mig_size*mig_nz_rotate); 
      wld_s=allocatec(MAX(mig_size,mig_nx*mig_ny_rotate)); wld_r=allocatec(MAX(mig_size,mig_nx*mig_ny_rotate));
      ep=allocatef(mig_size*mig_nz_rotate); dl=allocatef(mig_size*mig_nz_rotate);
      printf("nz_s_rotate=%d\n",rotate_par.nz_s_rotate);
      //if (0.0!=rotate_angle){
        wld_s_z0=allocatec(mig_size*rotate_par.nz_s_rotate); wld_r_z0=allocatec(mig_size*rotate_par.nz_s_rotate);
      //}
      image=allocatef(image_size); vector_value_f(image,0.0,image_size);

      wave.n[1]=mig_nx_rotate; wave.o[0]=0.0; wave.n[0]=mig_nz_rotate;
      if (rotate_par.rotate_angle>=0.0) wave.o[1]=mig_min_x; else wave.o[1]=mig_min_x+(mig_nx-1)*mig_dx;
      velocity_read(velocity,vel,wave,rotate_par.rotate_angle,rotate_par.ix_rotate_center,"Velocity"); veltranspose(vel,mig_ny_rotate,mig_nx_rotate,mig_nz_rotate);
      for(iw=0;iw<data_par.nw; iw++){ 
        w=iw*data_par.dw+data_par.ow;
        printf("iw=%d  w=%f\n",iw,w);
        
        shot_plane_source(mig_min_x,mig_min_y,mig_nx,mig_ny,iw,i_sx,i_sy,wld_s,data_par,plane_par);
        shot_plane_receiver(mig_min_x,mig_min_y,mig_nx,mig_ny,iw,i_sx,i_sy,wld_r,data_par,plane_par);
        //if (0.0!=rotate_angle){
          shot_rotate_data_rotate(wld_r,wld_r_z0,&rotate_par);
          shot_rotate_data_rotate(wld_s,wld_s_z0,&rotate_par);
          vector_value_c(wld_s,cmplx(0.0,0.0),mig_size); 
          vector_value_c(wld_r,cmplx(0.0,0.0),mig_size); //wld_r=0.0  wld_s=0.0
        //}
        for(iz=0;iz<mig_nz_rotate;iz++){
          image_z=image+mig_size*image_par.nhy*image_par.nhx*iz;
          //if (0.0!=rotate_angle){
          if (iz <rotate_par.nz_s_rotate){
             for(ix=0;ix<mig_nx_rotate;ix++) wld_s[ix]+=wld_s_z0[i2(iz,ix,mig_nx_rotate)];  //wld_s+=wld_s_z0(iz,:,:);
             for(ix=0;ix<mig_nx_rotate;ix++) wld_r[ix]+=wld_r_z0[i2(iz,ix,mig_nx_rotate)];  //wld_r+=wld_r_z0(iz,:,:);
          }
          if ( 0.0!=rotate_angle){
            shot_rotate_get_bx_ex_iz(&dkx,&bx,&ex,iz,10,rotate_par);
            nx_iz=ex-bx+1;  ker_par.dkx=dkx; ker_par.nx=nx_iz;
            v_z=vel+iz*mig_size+bx; wld_s_tilt=wld_s+bx; wld_r_tilt=wld_r+bx; image_z+=image_par.nhy*image_par.nhx*bx;
            ep_z=ep+iz*mig_size+bx; dl_z=dl+iz*mig_size+bx; 
          }
          //}
          else{
            v_z=vel+iz*mig_size; wld_s_tilt=wld_s; wld_r_tilt=wld_r; 
            nx_iz=mig_nx_rotate;
          }
          shot_ker_depth_onestep(wld_s_tilt,v_z,w,-1,ker_par);
          shot_ker_depth_onestep(wld_r_tilt,v_z,w,+1,ker_par); 
          shot_image_cross_tilt(wld_s_tilt,wld_r_tilt,image_z,mig_nx_rotate,mig_ny_rotate,image_par,nx_iz);
        }
      }
      /*
      if( i_sx==0){
        rotimage.ndims=2;
        rotimage.o[0]=0.0; rotimage.d[0]=mig_dx; rotimage.n[0]=mig_nx_rotate;
        rotimage.o[1]=0.0; rotimage.d[1]=mig_dz; rotimage.n[1]=mig_nz_rotate;
        stat=file_open_output("Imagerot",rotimage);
        if (srite("Imagerot",vel,mig_nx_rotate*mig_nz_rotate*4)!=mig_nx_rotate*mig_nz_rotate*4) seperr("err image r");
        //if (srite("Imagerot",image,mig_nx_rotate*mig_nz_rotate*4)!=mig_nx_rotate*mig_nz_rotate*4) seperr("err image r");
      }
      */
      free(wld_s); free(wld_r);free(vel); free(ep); free(dl); 
      if (0.0!=rotate_angle){
        free(wld_s_z0); free(wld_r_z0); 
      }





      image_rotate=allocatef(mig_nx*mig_nz);  shot_rotate_image_rotate(image,image_rotate,rotate_par);  free(image); //rotate image 
      shot_image_stack(image_rotate,mig_min_x,mig_min_y,mig_nx,mig_ny,"Image_hx",image_par); free(image_rotate); // rite image
      shot_rotate_release(&rotate_par);
      shot_ker_release(&ker_par);
    }
  }
  shot_data_release(&data_par);
  return 0;
}

void getimagez(float *image,float *image_z,int iz,int ny,int nx,int nhy,int nhx,int nz){
  int nimage[4],nimagez[3];
  int iy,ix,ihy,ihx;
  d5(nx,nhy,nhx,nz,nimage); d4(nx,nhy,nhx,nimagez);
  for (iy=0;iy<ny;iy++)
    for(ix=0;ix<nx;ix++)
      for(ihy=0;ihy<nhy;ihy++)
        for(ihx=0;ihx<nhx;ihx++)
           image[i5(iy,ix,ihy,ihx,iz,nimage)]+=image_z[i4(iy,ix,ihy,ihx,nimagez)];
}

void mig_grab_par(int *op,int *mig_nz,float *mig_dz,float *fd_a,float *fd_b,float *trick,int *tilt){
  if (0==getch("operator","d",op))    *op=1;
  if (0==getch("mig_dz","f",mig_dz))  *mig_dz=4.0;
  if (0==getch("mig_nz","d",mig_nz))  *mig_nz=1000;
  if (0==getch("fd_a","f",fd_a))      *fd_a=0.5;
  if (0==getch("fd_b","f",fd_b))      *fd_b=0.0;
  if (0==getch("fd_trick","f",trick)) *trick=1.0/6.0;
  if (0==getch("tilt","d",tilt)) *tilt=1;
}

void veltranspose(float *vel,int ny,int nx,int nz)
{
  float *tmpvel=allocatef(ny*nx*nz);
  int nvel[2],ntmpvel[2];
  int i,iz,iy,ix;
  d3(ny,nx,nvel); d3(nx,nz,ntmpvel);
  for(i=0;i<ny*nx*nz;i++)
    tmpvel[i]=vel[i];
  for(iz=0;iz<nz;iz++)
    for(iy=0;iy<ny;iy++)
      for(ix=0;ix<nx;ix++)
         vel[i3(iz,iy,ix,nvel)]=tmpvel[i3(iy,ix,iz,ntmpvel)];
  free(tmpvel); 
}
