module data_mod
use sep
use sep3d_struct_mod
implicit none


type data_par_type
  real :: dhx,dhy,dw,dsx,dsy
  real :: ohx,ohy,ow,osx,osy
  integer :: nhx,nhy,nw,nsx,nsy
  integer ::MPI_size
end type data_par_type

type plane_par_type
  real ::plane_os_x,plane_ds_x,plane_os_y,plane_ds_y
  integer :: plane_ns_x,plane_ns_y
  integer :: MPI_size
end type plane_par_type

logical :: migration
type(sep3d),private :: data,wavlt
type(data_par_type),private :: data_par
type(plane_par_type),private:: plane_par
complex,allocatable,private :: wavelet(:)

contains
subroutine shot2plane_init(ns,nw,dx,dy,dw,w0,i_node)
integer ::ns,nw,i_node
real :: dx,dy,dw,w0

real :: wlt_o, wlt_d
integer :: wlt_n
integer :: size_real,size_logical,size_int
integer,external :: auxin,sreed
integer :: ierr

if (i_node==0) then
  call init_sep3d("Data",data,"INPUT")
  data_par%dw =data%d(3); data_par%ow =data%o(3); data_par%nw =data%n(3)
  data_par%dhx=data%d(1); data_par%ohx=data%o(1); data_par%nhx=data%n(1)
  data_par%dhy=data%d(2); data_par%ohy=data%o(2); data_par%nhy=data%n(2)
  data_par%dsx=data%d(4); data_par%osx=data%o(4); data_par%nsx=data%n(4)
  data_par%dsy=data%d(5); data_par%osy=data%o(5); data_par%nsy=data%n(5)
  data_par%dsy=10.0;      data_par%osy=0.0;   data_par%nsy=1
end if  
ns=data_par%nsx*data_par%nsy
dx=data_par%dhx;  dy=data_par%dhy
dw=data_par%dw;   w0=data_par%ow; nw=data_par%nw
end subroutine shot2plane_init


subroutine plane_write()
type(sep3d):: plane_data
real :: opx,dpx,opy,dpy
integer :: npx,npy
real :: x_min,y_min,x_max,y_max,tmp_x_min,tmp_y_min
integer :: plane_nx,plane_ny,tmp_plane_nx,tmp_plane_ny
complex,allocatable :: wld(:,:,:)
integer :: ipx,ipy
real    :: px,py
real ::plane_osx,plane_dsx,plane_osy,plane_dsy
integer ::plane_nsx,plane_nsy
integer :: blocksize
integer,external :: srite

call from_param('opx',opx,0.0)
call from_param('dpx',dpx,1.0)
call from_param('npx',npx,30)
call from_param('opy',opy,0.0)
call from_param('dpy',dpy,1.0)
call from_param('npy',npy,1)


plane_osx=data_par%osx; plane_dsx=data_par%dsx; plane_nsx=data_par%nsx
plane_osy=data_par%osy; plane_dsy=data_par%dsy; plane_nsy=data_par%nsy
tmp_x_min=min(data_par%osx,data_par%osx+data_par%dsx*(data_par%nsx-1))&
+min(data_par%ohx+data_par%dhx*(data_par%nhx-1),data_par%ohx)
!tmp_x_min=data_par%ohx
write(*,*) "tmp_x_min",tmp_x_min
tmp_y_min=min(data_par%osy+data_par%dsy*(data_par%nsy-1),data_par%osy)&
+min(data_par%ohy+data_par%dhy*(data_par%nhy-1),data_par%ohy)
!tmp_y_min=data_par%ohy
call from_param('plane_x_min',x_min,tmp_x_min)
call from_param('plane_y_min',y_min,tmp_y_min)
x_max=max(data_par%osx,data_par%osx+(data_par%nsx-1)*data_par%dsx)&
+max(data_par%ohx,data_par%ohx+(data_par%nhx-1)*data_par%dhx)
!x_max=data_par%ohx+(data_par%nhx-1)*data_par%dhx
write(*,*) "x_max",x_max
y_max=max(data_par%osy,data_par%osy+(data_par%nsy-1)*data_par%dsy)&
+max(data_par%ohy,data_par%ohy+(data_par%nhy-1)*data_par%dhy)
!y_max=data_par%ohy+(data_par%nhy-1)*data_par%dhy

tmp_plane_nx=(x_max-tmp_x_min)/abs(data_par%dhx)+1
write(*,*) "tmp_palne_nx",tmp_plane_nx
tmp_plane_ny=(y_max-tmp_y_min)/abs(data_par%dhy)+1
call from_param('plane_nx',plane_nx,tmp_plane_nx)
call from_param('plane_ny',plane_ny,tmp_plane_ny)

call init_sep3d(plane_data, "OUTPUT","COMPLEX", "REGULAR", ctag="Plane")
call axis_allocate(plane_data,5)
plane_data%o(1)=x_min;       plane_data%d(1)=abs(data_par%dhx);plane_data%n(1)=plane_nx
plane_data%o(2)=y_min;       plane_data%d(2)=abs(data_par%dhy);plane_data%n(2)=plane_ny
plane_data%o(3)=data_par%ow; plane_data%d(3)=data_par%dw; plane_data%n(3)=data_par%nw
plane_data%o(4)=opx;         plane_data%d(4)=dpx;         plane_data%n(4)=npx
plane_data%o(5)=opy;         plane_data%d(5)=dpy;         plane_data%n(5)=npy
call sep3d_set_sep3d(plane_data)
call sep3d_write_description("Plane",plane_data)

call auxputch ("plane_os_x","f",plane_osx,"Plane")
call auxputch ("plane_ds_x","f",plane_dsx,"Plane")
call auxputch ("plane_ns_x","i",plane_nsx,"Plane")
call auxputch ("plane_os_y","f",plane_osy,"Plane")
call auxputch ("plane_ds_y","f",plane_dsy,"Plane")
call auxputch ("plane_ns_y","i",plane_nsy,"Plane")

allocate(wld(plane_nx,plane_ny,data_par%nw))
write(*,*) "f1",plane_nx,plane_ny,data_par%nw
blocksize=plane_nx*plane_ny*data_par%nw*8
do ipy=1,npy
  py=(ipy-1)*dpy+opy
  do ipx=1,npx
    !write(*,*) "f2",ipx
    px=(ipx-1)*dpx+opx
    wld=0.0
    call plane_transfer(x_min,y_min,plane_nx,plane_ny,px,py,wld,&
     plane_osx,plane_dsx,plane_nsx,plane_osy,plane_dsy,plane_nsy)
    !write(*,*) "f3"
    if (srite("Plane",wld,blocksize)/=blocksize ) call seperr("err write")
  end do
end do

deallocate(wld)
end subroutine plane_write



subroutine plane_transfer(min_x,min_y,nx,ny,px,py,wld,&
plane_osx,plane_dsx,plane_nsx,plane_osy,plane_dsy,plane_nsy)
real :: min_x,min_y,px,py
integer :: nx,ny
complex :: wld(:,:,:)
real :: plane_osx,plane_dsx,plane_osy,plane_dsy
integer :: plane_nsx,plane_nsy

complex,allocatable:: tmp_wld(:,:,:)


real :: x_dt,y_dt,t_y,t_x
integer :: blocksize,i_block
integer :: plane_isy,plane_isx,isx,isy,iw,i_ox,i_oy
integer :: i_ox_min, i_ox_max,i_oy_min,i_oy_max,i_x,i_y
real :: plane_sy,shot_oy,plane_sx,shot_ox,w
integer, external:: sreed,sseek_block 

! plane_os_x: first point shot to construct plane source

x_dt=plane_dsx*px
y_dt=plane_dsy*py

write(0,*) "x_dt",x_dt,nx,ny
allocate(tmp_wld(data_par%nhx,data_par%nhy,data_par%nw))

blocksize=8*data_par%nhx*data_par%nhy*data_par%nw
!write(*,*)"plane_ns_y",plane_nsy,plane_nsx
do plane_isy=1,plane_nsy
  t_y=(plane_isy-1)*y_dt
  plane_sy=(plane_isy-1)*plane_dsy+plane_osy
  isy=(plane_sy-data_par%osy)/data_par%dsy+1 ! index in shot gather data
  shot_oy=plane_sy+data_par%ohy
!  shot_oy=data_par%ohy
  i_oy=(shot_oy-min_y)/data_par%dhy+1    ! index in plane wave
  !write(*,*)"i_oy", i_oy
  do plane_isx=1,plane_nsx
    plane_sx=(plane_isx-1)*plane_dsx+plane_osx  
    isx=(plane_sx-data_par%osx)/data_par%dsx+1
    t_x=(plane_isx-1)*x_dt
    shot_ox=plane_sx+data_par%ohx
!    shot_ox=data_par%ohx
    i_ox=(shot_ox-min_x)/abs(data_par%dhx)+1
    !!!Read a shot at :isx,isy  to tmp_wld(x,y,w)
    i_block=(isy-1)*data_par%nsx+isx-1
    if ( sseek_block("Data",i_block,blocksize,0) /=i_block) call seperr("plane source data seek wrong")
    if ( sreed("Data",tmp_wld,blocksize )/= blocksize) call seperr("plane source data read wrong")
    !!! wld(i_ox:i_ox+nhx,i_oy:i_oy+nhy,w)+=tmp_wld*cmplx(cos(w*t),sin(w*t))
    i_ox_min=min(i_ox,i_ox+sign(data_par%dhx)*(data_par%nhx-1))
    i_ox_max=max(i_ox,i_ox+sign(data_par%dhx)*(data_par%nhx-1))
    i_oy_min=min(i_oy,i_oy+sign(data_par%dhy)*(data_par%nhy-1))
    i_oy_max=max(i_oy,i_oy+sign(data_par%dhy)*(data_par%nhy-1))
    do iw=1,data_par%nw
      w=(iw-1)*data_par%dw+data_par%ow
      if (data_par%dhx >0) then
        wld(i_ox_min:i_ox_max,i_oy_min:i_oy_max,iw)=&
          wld(i_ox_min:i_ox_max,i_oy_min:i_oy_max,iw) +&
          tmp_wld(:,:,iw)*cmplx(cos(w*t_x),sin(w*t_x))
      else
        do i_y=1,data_par%nhy
          do i_x=1,data_par%nhx
            wld(i_ox-i_x+1,i_y+i_oy-1,iw)=wld(i_x-i_x+1,i_y+i_oy-1,iw)+&
            tmp_wld(i_x,i_y,iw)*cmplx(cos(w*t_x),sin(w*t_x))
          end do
        end do
      end if
    end do
  end do
end do
deallocate(tmp_wld)
end subroutine plane_transfer

integer function sign(a)
real :: a
if (a>0) sign=1
if (a==0) sign=0
if (a <0) sign=-1
return
end function sign


end module data_mod
